import drjit as dr
import pytest
import importlib

def prepare(name):
    if 'cuda' in name:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in name:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    elif 'packet' in name and not hasattr(dr, 'packet'):
        pytest.skip('Packet mode is unsupported')
    return importlib.import_module(name)

@pytest.mark.parametrize("package", ["drjit.cuda", "drjit.cuda.ad",
                                     "drjit.llvm", "drjit.llvm.ad"])
def test_roundtrip_dlpack_all(package):
    package = prepare(package)
    Float, Array3f = package.Float, package.Array3f
    a1 = Array3f(
        dr.arange(Float, 10),
        dr.arange(Float, 10)+100,
        dr.arange(Float, 10)+1000,
    )
    a2 = a1.__dlpack__()
    a3 = Array3f(a2)
    assert a1 == a3
    assert a1.x == Float(a1.x.__dlpack__())


@pytest.mark.parametrize("package", ["drjit.cuda", "drjit.cuda.ad",
                                     "drjit.llvm", "drjit.llvm.ad",
                                     "drjit.packet"])
def test_roundtrip_numpy_all(package):
    pytest.importorskip("numpy")
    package = prepare(package)
    Float, Array3f = package.Float, package.Array3f
    a1 = Array3f(
        dr.arange(Float, 10),
        dr.arange(Float, 10)+100,
        dr.arange(Float, 10)+1000,
    )
    a2 = a1.numpy()
    a3 = Array3f(a2)
    assert a1 == a3
    assert a1.x == Float(a1.x.numpy())

    if Float.IsDynamic:
        import numpy as np
        assert len(np.array(Float())) == 0
        assert len(Float(np.array([], dtype=np.float32))) == 0


def test_roundtrip_pytorch_cuda():
    pytest.importorskip("torch")
    prepare("drjit.cuda.ad")
    from drjit.cuda.ad import Float, Array3f
    a1 = Array3f(
        dr.arange(Float, 10),
        dr.arange(Float, 10)+100,
        dr.arange(Float, 10)+1000,
    )
    a2 = a1.torch()
    a3 = Array3f(a2)
    assert a1 == a3
    assert a1.x == Float(a1.x.torch())


def test_roundtrip_pytorch_llvm():
    pytest.importorskip("torch")
    prepare("drjit.llvm.ad")
    from drjit.llvm.ad import Float, Array3f
    a1 = Array3f(
        dr.arange(Float, 10),
        dr.arange(Float, 10)+100,
        dr.arange(Float, 10)+1000,
    )
    a2 = a1.torch()
    a3 = Array3f(a2)
    assert a1 == a3
    assert a1.x == Float(a1.x.torch())


def test_roundtrip_pytorch_jax():
    pytest.importorskip("jax")

    try:
        from jax.lib import xla_bridge
        xla_bridge.get_backend('gpu')
    except:
        pytest.skip('Backend gpu failed to initialize on JAX')

    prepare("drjit.cuda.ad")
    from drjit.cuda.ad import Float, Array3f
    a1 = Array3f(
        dr.arange(Float, 10),
        dr.arange(Float, 10)+100,
        dr.arange(Float, 10)+1000,
    )
    a2 = a1.jax()
    a3 = Array3f(a2)
    assert a1 == a3
    assert a1.x == Float(a1.x.jax())


def test_matrix_numpy_construction():
    pytest.importorskip("numpy")
    import numpy as np

    m_py = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    m_np = np.array(m_py)
    m_ek1 = dr.scalar.Matrix3f(m_py)
    m_ek2 = dr.scalar.Matrix3f(m_np)

    assert dr.allclose(m_ek1, m_ek2)
    assert dr.allclose(m_ek1, m_py)
    assert dr.allclose(m_ek1, m_np)


def test_matrix_3_to_4_conversion():
    m3 = dr.scalar.Matrix3f(*range(1, 10))
    m4 = dr.scalar.Matrix4f(m3)

    assert dr.allclose(m4, [[1, 2, 3, 0], [4, 5, 6, 0], [7, 8, 9, 0], [0, 0, 0, 1]])
    assert dr.allclose(m3, dr.scalar.Matrix3f(m4))



import drjit as dr
import pytest
import gc


def get_class(name):
    """Resolve a package+class name into the corresponding type"""
    if 'cuda' in name:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in name:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    elif 'packet' in name and not hasattr(dr, 'packet'):
        pytest.skip('Packet mode is unsupported')

    name = name.split('.')
    value = __import__(".".join(name[:-1]))
    for item in name[1:]:
        value = getattr(value, item)
    return value


def test01_construct():
    a = dr.scalar.Array3f(1, 2, 3)
    assert(repr(a) == "[1.0, 2.0, 3.0]")
    a = dr.scalar.Array3f([1, 2, 3])
    assert(repr(a) == "[1.0, 2.0, 3.0]")
    a = dr.scalar.Array3f((1, 2, 3))
    assert(repr(a) == "[1.0, 2.0, 3.0]")
    a = dr.scalar.Array3f(4)
    assert(repr(a) == "[4.0, 4.0, 4.0]")

    with pytest.raises(TypeError) as ei:
        a = dr.scalar.Array3f("asdf")

    assert "Array3f constructor expects: 1 or 3 values of type \"float\", a " \
        "matching list/tuple, or a NumPy/PyTorch/TF/Jax array." in str(ei.value)


@pytest.mark.parametrize("cname", ["drjit.scalar.Array1f",
                                   "drjit.scalar.Array1i",
                                   "drjit.packet.Float",
                                   "drjit.packet.Int",
                                   "drjit.cuda.Float",
                                   "drjit.cuda.Int",
                                   "drjit.cuda.ad.Float",
                                   "drjit.cuda.ad.Int",
                                   "drjit.llvm.Float",
                                   "drjit.llvm.Int",
                                   "drjit.llvm.ad.Float",
                                   "drjit.llvm.ad.Int"
                                   ])
def test02_basic_ops(cname):
    t = get_class(cname)

    # Test basic arithmetic operations
    import drjit.router as r
    nc = r._entry_evals
    v1 = t(4)
    v2 = t(2)
    assert v1[0] == 4
    assert v2[0] == 2
    v3 = v1 + v2
    assert v3[0] == 6
    v3 = v1 - v2
    assert v3[0] == 2
    v3 = v1 * v2
    assert v3[0] == 8

    if dr.scalar_t(t) is int:
        v3 = v1 // v2
        assert v3[0] == 2
        with pytest.raises(TypeError):
            v3 = v1 / v2
    else:
        v3 = v1 / v2
        assert v3[0] == 2.0
        with pytest.raises(TypeError):
            v3 = v1 // v2

    # Make sure that in-place ops are truly in-place
    v = t(1)
    v_id = id(v)

    if dr.scalar_t(t) is int:
        v //= v
        v >>= 0
        v <<= 0
        v %= v
    else:
        v /= v

    v += v
    v -= v
    v *= v

    assert v_id == id(v)

    # We now exactly how many times Array.entry()/set_entry()
    # should have been called during all of the above
    entry_evals = 6
    if 'scalar' in cname:
        entry_evals = 39 if dr.scalar_t(t) is int else 30

    assert r._entry_evals == nc + entry_evals
    assert not t.IsDynamic == hasattr(v3, 'x')


@pytest.mark.parametrize("pkg", ['drjit.packet', 'drjit.cuda', 'drjit.llvm'])
def test03_type_promotion(pkg):
    m = get_class(pkg)

    assert type(m.Array1b() | m.Array1b()) is m.Array1b
    assert type(m.Array1b() ^ m.Array1b()) is m.Array1b
    assert type(m.Array1b() & m.Array1b()) is m.Array1b

    with pytest.raises(dr.Exception) as ei:
        _ = m.Array1b() + m.Array1b()
    assert "add(): requires arithmetic operands!" in str(ei.value)

    assert type(m.Array1i() + m.Array1i()) is m.Array1i
    assert type(m.Array1i() + m.Array1u()) is m.Array1u
    assert type(m.Array1i() + m.Array1i64()) is m.Array1i64
    assert type(m.Array1i() + m.Array1u64()) is m.Array1u64
    assert type(m.Array1i() + m.Array1f()) is m.Array1f
    assert type(m.Array1i() + m.Array1f64()) is m.Array1f64

    with pytest.raises(dr.Exception) as ei:
        _ = dr.sqrt(m.Array1i())
    assert "sqrt(): requires floating point operands!" in str(ei.value)

    assert type(m.Array1f() + m.Array1f()) is m.Array1f
    assert type(m.Array1f() + m.Array1u()) is m.Array1f
    assert type(m.Array1f() + m.Array1i64()) is m.Array1f
    assert type(m.Array1f() + m.Array1u64()) is m.Array1f
    assert type(m.Array1f() + m.Array1f()) is m.Array1f
    assert type(m.Array1f() + m.Array1f64()) is m.Array1f64

    assert type(m.Array1f() | m.Array1b()) is m.Array1f
    assert type(m.Array1f() ^ m.Array1b()) is m.Array1f
    assert type(m.Array1f() & m.Array1b()) is m.Array1f


# Run various standard operations on a 3D array
def test04_operators():
    I3 = dr.scalar.Array3i
    F3 = dr.scalar.Array3f

    assert(I3(1, 2, 3) + I3(0, 1, 0) == I3(1, 3, 3))
    assert(I3(1, 2, 3) - I3(0, 1, 0) == I3(1, 1, 3))
    assert(I3(1, 2, 3) * I3(0, 1, 0) == I3(0, 2, 0))
    assert(I3(1, 2, 3) // I3(2, 2, 2) == I3(0, 1, 1))
    assert(I3(1, 2, 3) % I3(2, 2, 2) == I3(1, 0, 1))
    assert(I3(1, 2, 3) << 1 == I3(2, 4, 6))
    assert(I3(1, 2, 3) >> 1 == I3(0, 1, 1))
    assert(I3(1, 2, 3) & 1 == I3(1, 0, 1))
    assert(I3(1, 2, 3) | 1 == I3(1, 3, 3))
    assert(I3(1, 2, 3) ^ 1 == I3(0, 3, 2))
    assert(-I3(1, 2, 3) == I3(-1, -2, -3))
    assert(~I3(1, 2, 3) == I3(-2, -3, -4))
    assert(dr.abs(I3(1, -2, 3)) == I3(1, 2, 3))
    assert(abs(I3(1, -2, 3)) == I3(1, 2, 3))
    assert(dr.abs(I3(1, -2, 3)) == I3(1, 2, 3))
    assert(dr.fma(F3(1, 2, 3), F3(2, 3, 1), F3(1, 1, 1)) == F3(3, 7, 4))


def all_arrays(cond=lambda x: True):
    a = list(dr.scalar.__dict__.items())
    if hasattr(dr, "packet"):
        a += dr.packet.__dict__.items()
    if dr.has_backend(dr.JitBackend.CUDA):
        a += dr.cuda.__dict__.items()
    if dr.has_backend(dr.JitBackend.LLVM):
        a += dr.llvm.__dict__.items()
    return [v for k, v in a if isinstance(v, type) and cond(v)
            and not dr.is_special_v(v)
            and not dr.is_tensor_v(v)
            and not dr.is_texture_v(v)
            and not dr.depth_v(v) >= 3
            and not (dr.depth_v(v) >= 2 and 'scalar' in v.__module__)]


# Run various standard operations on *every available* type
@pytest.mark.parametrize("t", all_arrays())
def test05_scalar(t):
    if not dr.is_array_v(t) or dr.size_v(t) == 0:
        return
    get_class(t.__module__)

    if dr.is_mask_v(t):
        assert dr.all_nested(t(True))
        assert dr.any_nested(t(True))
        assert dr.none_nested(t(False))
        assert dr.all_nested(t(False) ^ t(True))
        assert dr.all_nested(dr.eq(t(False), t(False)))
        assert dr.none_nested(dr.eq(t(True), t(False)))

    if dr.is_arithmetic_v(t):
        assert t(1) + t(1) == t(2)
        assert t(3) - t(1) == t(2)
        assert t(2) * t(2) == t(4)
        assert dr.minimum(t(2), t(3)) == t(2)
        assert dr.maximum(t(2), t(3)) == t(3)

        if dr.is_signed_v(t):
            assert t(2) * t(-2) == t(-4)
            assert dr.abs(t(-2)) == t(2)

        if dr.is_integral_v(t):
            assert t(6) // t(2) == t(3)
            assert t(7) % t(2) == t(1)
            assert t(7) >> 1 == t(3)
            assert t(7) << 1 == t(14)
            assert t(1) | t(2) == t(3)
            assert t(1) ^ t(3) == t(2)
            assert t(1) & t(3) == t(1)
        else:
            assert t(6) / t(2) == t(3)
            assert dr.sqrt(t(4)) == t(2)
            assert dr.fma(t(1), t(2), t(3)) == t(5)
            assert (t(1) & True) == t(1)
            assert (t(1) & False) == t(0)
            assert (t(1) | False) == t(1)

        assert dr.all_nested(t(3) > t(2))
        assert dr.all_nested(dr.eq(t(2), t(2)))
        assert dr.all_nested(dr.neq(t(3), t(2)))
        assert dr.all_nested(t(1) >= t(1))
        assert dr.all_nested(t(2) < t(3))
        assert dr.all_nested(t(1) <= t(1))
        assert dr.select(dr.eq(t(2), t(2)), t(4), t(5)) == t(4)
        assert dr.select(dr.eq(t(3), t(2)), t(4), t(5)) == t(5)
        t2 = t(2)
        assert dr.sum(t2) == t.Value(2 * len(t2))
        assert dr.dot(t2, t2) == t.Value(4 * len(t2))

        value = t(1)
        value[dr.eq(value, t(1))] = t(2)
        value[dr.eq(value, t(3))] = t(5)
        assert value == t(2)


def test06_reinterpret_cast():
    I3 = dr.scalar.Array3i
    F3 = dr.scalar.Array3f
    B3 = dr.scalar.Array3b
    LI3 = dr.scalar.Array3i64
    LF3 = dr.scalar.Array3f64

    assert dr.mask_t(I3) is B3
    assert dr.scalar_t(F3) is float
    assert dr.scalar_t(I3) is int
    assert dr.float_array_t(I3) is F3
    assert dr.int_array_t(F3) is I3
    assert dr.float_array_t(LI3) is LF3
    assert dr.int_array_t(LF3) is LI3

    assert dr.reinterpret_array_v(I3, F3(1)).x == 0x3f800000
    assert dr.reinterpret_array_v(F3, I3(0x3f800000)).x == 1.0


@pytest.mark.parametrize("pkg", ['drjit.cuda', 'drjit.llvm'])
def test07_gather_ravel_unravel(pkg):
    pkg = get_class(pkg)
    str_1 = '[[0.0, 1.0, 2.0],\n [3.0, 4.0, 5.0],\n [6.0, 7.0, 8.0],\n' \
        ' [9.0, 10.0, 11.0]]'
    str_2 = '[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0]'
    a = dr.arange(pkg.Float, 100)
    b = dr.gather(pkg.Array3f, a, pkg.UInt(0, 1, 2, 3))
    assert repr(b) == str_1
    c = dr.ravel(b)
    assert repr(c) == str_2
    d = dr.unravel(pkg.Array3f, c)
    assert repr(d) == str_1


@pytest.mark.parametrize("t", all_arrays(lambda a:
                                         getattr(a, 'IsFloat', False)))
def test07_sincos(t):
    def poly2(x, c0, c1, c2):
        x2 = dr.sqr(x)
        return dr.fma(x2, c2, dr.fma(x, c1, c0))

    def sincos(x):
        Float = type(x)
        Int = dr.int_array_t(Float)

        xa = dr.abs(x)

        j = Int(xa * 1.2732395447351626862)

        j = (j + Int(1)) & ~Int(1)

        y = Float(j)

        Shift = Float.Type.Size * 8 - 3

        sign_sin = dr.reinterpret_array_v(Float, j << Shift) ^ x
        sign_cos = dr.reinterpret_array_v(Float, (~(j - Int(2)) << Shift))

        y = xa - y * 0.78515625 \
               - y * 2.4187564849853515625e-4 \
               - y * 3.77489497744594108e-8

        z = y * y
        z |= dr.eq(xa, dr.inf)

        s = poly2(z, -1.6666654611e-1,
                  8.3321608736e-3,
                  -1.9515295891e-4) * z

        c = poly2(z, 4.166664568298827e-2,
                  -1.388731625493765e-3,
                  2.443315711809948e-5) * z

        s = dr.fma(s, y, y)
        c = dr.fma(c, z, dr.fma(z, -0.5, 1))

        polymask = dr.eq(j & Int(2), dr.zeros(Int))

        return (
            dr.mulsign(dr.select(polymask, s, c), sign_sin),
            dr.mulsign(dr.select(polymask, c, s), sign_cos)
        )

    get_class(t.__module__)
    s, c = sincos(t(1))
    if t.Size != 0:
        assert dr.allclose(s**2 + c**2, 1)


@pytest.mark.parametrize("cname", ["drjit.packet.Int",
                                   "drjit.packet.Int64",
                                   "drjit.packet.UInt",
                                   "drjit.packet.UInt64",
                                   "drjit.cuda.Int",
                                   "drjit.cuda.Int64",
                                   "drjit.cuda.UInt",
                                   "drjit.cuda.UInt64",
                                   "drjit.llvm.Int",
                                   "drjit.llvm.Int64",
                                   "drjit.llvm.UInt",
                                   "drjit.llvm.UInt64"])
def test08_divmod(cname):
    t = get_class(cname)

    index = dr.arange(t, 10000000)
    index[index < len(index) // 2] = -index
    index *= 256203161

    for i in range(1, 100):
        assert index // i == index // dr.opaque(t, i, 1)
        assert index % i == index % dr.opaque(t, i, 1)

    if t.IsSigned:
        for i in range(1, 100):
            assert index // -i == index // dr.opaque(t, -i, 1)
            assert index % -i == index % dr.opaque(t, -i, 1)


@pytest.mark.parametrize("cname", ["drjit.cuda.Float", "drjit.llvm.Float"])
def test09_repeat_tile(cname):
    t = get_class(cname)
    a3 = get_class(cname.replace('Float', 'Array3f'))
    vec = t([1, 2, 3])
    tiled = t([1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3, 1, 2, 3])
    reptd = t([1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3])

    assert dr.tile(vec, 5) == tiled
    assert dr.tile(a3(vec, vec + 1, vec + 2), 5) == \
        a3(tiled, tiled + 1, tiled + 2)

    assert dr.repeat(vec, 5) == reptd
    assert dr.repeat(a3(vec, vec + 1, vec + 2), 5) == \
        a3(reptd, reptd + 1, reptd + 2)


@pytest.mark.parametrize("cname", ["drjit.cuda.Int", "drjit.llvm.Int"])
def test10_meshgrid(cname):
    np = pytest.importorskip("numpy")

    Int = get_class(cname)

    assert dr.meshgrid() == ()

    assert dr.meshgrid(Int(1, 2), indexing='ij') == Int(1, 2)
    assert dr.meshgrid(Int(1, 2), indexing='xy') == Int(1, 2)

    assert dr.meshgrid(Int(1, 2), Int(3, 4, 5)) == \
        (Int(1, 2, 1, 2, 1, 2), Int(3, 3, 4, 4, 5, 5))
    assert dr.meshgrid(Int(1, 2), Int(3, 4, 5), indexing='ij') == \
        (Int(1, 1, 1, 2, 2, 2), Int(3, 4, 5, 3, 4, 5))

    assert dr.meshgrid(Int(1, 2), Int(3, 4, 5), Int(5, 6), indexing='xy') == \
       (Int(1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2),
        Int(3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5),
        Int(5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6))
    assert dr.meshgrid(Int(1, 2), Int(3, 4, 5), Int(5, 6), indexing='ij') == \
       (Int(1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2),
        Int(3, 3, 4, 4, 5, 5, 3, 3, 4, 4, 5, 5),
        Int(5, 6, 5, 6, 5, 6, 5, 6, 5, 6, 5, 6))

    # Ensure consistency with NumPy
    a, b = dr.meshgrid(Int(1, 2), Int(3, 4, 5))
    a_np, b_np = np.meshgrid((1, 2), (3, 4, 5))
    assert a == a_np.ravel()
    assert b == b_np.ravel()
    a, b = dr.meshgrid(Int(1, 2), Int(3, 4, 5), indexing='ij')
    a_np, b_np = np.meshgrid((1, 2), (3, 4, 5), indexing='ij')
    assert a == a_np.ravel()
    assert b == b_np.ravel()

    a, b, c = dr.meshgrid(Int(1, 2), Int(3, 4, 5), Int(5, 6))
    a_np, b_np, c_np = np.meshgrid((1, 2), (3, 4, 5), Int(5, 6))
    assert a == a_np.ravel()
    assert b == b_np.ravel()
    assert c == c_np.ravel()
    a, b, c = dr.meshgrid(Int(1, 2), Int(3, 4, 5), Int(5, 6), indexing='ij')
    a_np, b_np, c_np = np.meshgrid((1, 2), (3, 4, 5), Int(5, 6), indexing='ij')
    assert a == a_np.ravel()
    assert b == b_np.ravel()
    assert c == c_np.ravel()



@pytest.mark.parametrize("cname", ["drjit.cuda.Int", "drjit.llvm.Int"])
def test11_block_sum(cname):
    Int = get_class(cname)
    assert dr.block_sum(Int(1, 2, 3, 4), 2) == Int(3, 7)


@pytest.mark.parametrize("cname", ["drjit.cuda.Float", "drjit.llvm.Float"])
def test12_binary_search(cname):
    np = pytest.importorskip("numpy")

    t = get_class(cname)

    data_np = np.float32(np.sort(np.random.normal(size=10000)))
    search_np = np.float32(np.random.normal(size=10000))
    data = t(data_np)
    search = t(search_np)

    index = dr.binary_search(
        0, len(data) - 1,
        lambda index: dr.gather(t, data, index) < search
    )

    value = dr.gather(t, data, index)
    cond = dr.eq(index, len(data)-1) | (value >= search)
    assert dr.all(cond)


def test13_slice_setitem():
    a = dr.zeros(dr.scalar.ArrayXf, 5)
    a[2] = 1.0
    assert dr.allclose(a, [0, 0, 1, 0, 0])
    a[2:] = [2.0, 1.0, 1.0]
    assert dr.allclose(a, [0, 0, 2, 1, 1])
    a[:] = 0.0
    assert dr.allclose(a, [0, 0, 0, 0, 0])

    v = dr.scalar.Array3f(0)
    v[2] = 1.0
    assert dr.allclose(v, [0, 0, 1])
    v[1:] = 2.0
    assert dr.allclose(v, [0, 2, 2])

    m = dr.scalar.Matrix3f(0)
    m[1:, 1] = 1.0
    assert dr.allclose(m, [[0, 0, 0], [0, 1, 0], [0, 1, 0]])
    m[0, 1:] = 2.0
    assert dr.allclose(m, [[0, 2, 2], [0, 1, 0], [0, 1, 0]])
    m[1:, 1:] = 3.0
    assert dr.allclose(m, [[0, 2, 2], [0, 3, 3], [0, 3, 3]])


def test14_slice_getitem():
    a = dr.scalar.ArrayXf(*range(5))
    assert dr.allclose(a[1:4], [1, 2, 3])
    assert dr.allclose(a[:], [0, 1, 2, 3, 4])

    a = dr.scalar.ArrayXi(*range(5))
    assert dr.allclose(a[1:4], [1, 2, 3])
    assert dr.allclose(a[:], [0, 1, 2, 3, 4])


def test15_test_avx512_approx():
    Float = get_class('drjit.llvm.Float')

    x = dr.linspace(Float, 0, 10, 1000)
    o = dr.full(Float, 1, 1000)
    assert dr.allclose(dr.rsqrt(x), o / dr.sqrt(x), rtol=2e-7, atol=0)
    assert dr.allclose(dr.rcp(x), o / x, rtol=2e-7, atol=0)


@pytest.mark.parametrize("cname", ["drjit.cuda.PCG32", "drjit.llvm.PCG32"])
def test16_custom(cname):
    t = get_class(cname)

    v1 = dr.zeros(t, 100)
    v2 = dr.empty(t, 100)
    assert len(v1.state) == 100
    assert len(v2.inc) == 100

    v2.state = v1.state
    v1.state = dr.arange(type(v1.state), 100)
    v3 = dr.select(v1.state < 10, v1, v2)
    assert v3.state[3] == 3
    assert v3.state[11] == 0

    assert dr.width(v3) == 100
    v4 = dr.zeros(t, 1)
    dr.schedule(v4)
    dr.resize(v4, 200)
    assert dr.width(v4) == 200

    assert dr.width(v3) == 100
    v4 = dr.zeros(t, 1)
    dr.resize(v4, 200)
    assert dr.width(v4) == 200

    index = dr.arange(type(v1.state), 100)
    dr.scatter(v4, v1, index)
    v5 = dr.gather(t, v4, index)
    dr.eval(v5)
    assert v5.state == v1.state and v5.inc == v1.inc


def test17_opaque():
    Array3f = get_class('drjit.llvm.Array3f')

    v = dr.opaque(Array3f, 4.0)
    assert dr.width(v) == 1
    assert dr.allclose(v, 4.0)

    for i in range(len(v)):
        assert not v[i].is_literal_()
        assert v[i].is_evaluated_()

    v = dr.opaque(Array3f, 4.0, 1)
    assert dr.width(v) == 1
    assert dr.allclose(v, 4.0)

    v = dr.opaque(Array3f, 4.0, 10)
    assert dr.width(v) == 10
    assert dr.allclose(v, 4.0)


def test18_slice():
    Float   = get_class('drjit.llvm.Float')
    Array2f = get_class('drjit.llvm.Array2f')

    a = dr.arange(Float, 10)
    for i in range(10):
        assert dr.slice(a, i) == i

    a2 = Array2f(a, a * 10)
    for i in range(10):
        assert dr.slice(a2, i)[0] == i
        assert dr.slice(a2, i)[1] == i * 10

    a3 = (a, a * 10)
    for i in range(10):
        assert dr.slice(a3, i)[0] == i
        assert dr.slice(a3, i)[1] == i * 10


def test19_make_opaque():
    Float   = get_class('drjit.llvm.Float')
    Array3f = get_class('drjit.llvm.Array3f')
    TensorXf = get_class('drjit.llvm.TensorXf')

    a = Float(4.4)
    b = dr.full(Float, 3.3, 10)
    c = Array3f(2.2, 5.5, 6.6)
    t = dr.full(TensorXf, 4.4, (4, 4, 4))

    assert a.is_literal_()
    assert not a.is_evaluated_()
    assert b.is_literal_()
    assert not b.is_evaluated_()
    for i in range(len(c)):
        assert c[i].is_literal_()
        assert not c[i].is_evaluated_()
    assert t.array.is_literal_()
    assert not t.array.is_evaluated_()

    dr.make_opaque(a, b, c, t)

    assert not a.is_literal_()
    assert a.is_evaluated_()
    assert not b.is_literal_()
    assert b.is_evaluated_()
    for i in range(len(c)):
        assert not c[i].is_literal_()
        assert c[i].is_evaluated_()
    assert not t.array.is_literal_()
    assert t.array.is_evaluated_()


def test20_unsigned_negative():
    UInt32 = get_class('drjit.llvm.UInt32')

    a = UInt32(-1)
    assert a.is_literal_()
    assert a == 4294967295

    a = UInt32(1)
    b = -1
    c = a - b

    assert c.is_literal_()
    assert c == 2

@pytest.mark.parametrize("t", all_arrays())
def test21_masked_update(t):
    if dr.is_array_v(t) and not dr.is_mask_v(t):
        v1, v2 = t(1), t(5)
        v1[v1 < 2] = 3
        v2[v2 < 2] = 3

        assert v1 == t(3) and v2 == t(5)

        if t.IsJIT and t.Depth == 1:
            x = dr.arange(t, 100)
            x[x < 5] += 5
            y = dr.arange(t, 100)
            y = dr.select(y < 5, y + 5, y)
            assert x == y


@pytest.mark.parametrize("pkg", ['drjit.scalar', 'drjit.cuda', 'drjit.llvm'])
def test22_sh_eval(pkg):
    m = get_class(pkg)

    # From Mathematica
    ref = [
        0.2820947918,   -0.2611690283,    0.3917535424,  -0.1305845141,
        0.1560783472,   -0.4682350417,    0.2928635963,  -0.2341175208,
       -0.1170587604,    0.02252796895,   0.3310921732,  -0.5409527810,
        0.06411572364,  -0.2704763905,   -0.2483191299,   0.1239038292,
       -0.07663294720,   0.05418767663,   0.4730873479,  -0.4301013494,
       -0.1926808176,   -0.2150506747,   -0.3548155109,   0.2980322214,
       -0.02235127627,   0.03401106316,  -0.2037835426,   0.08939333855,
        0.5098360937,   -0.1642890433,   -0.3717265730,  -0.08214452164,
       -0.3823770702,    0.4916633621,   -0.05943686658, -0.03669614710,
        0.01095484717,   0.09832164158,  -0.3751138475,   0.1148149412,
        0.4035308753,    0.1617920839,   -0.4010373398,   0.08089604195,
       -0.3026481565,    0.6314821768,   -0.1094082055,  -0.1060838764,
        0.02912993451,  -0.01914767451,   0.03401803232,  0.1978196393,
       -0.5458487747,    0.1174915463,    0.1714596820,   0.4252852218,
       -0.2693765472,    0.2126426109,   -0.1285947615,   0.6462035049,
       -0.1592058926,   -0.2134369792,    0.09045704049, -0.001997419283,
        0.006375451838, -0.06329912897,   0.07377498039,  0.3173673505,
       -0.6530654560,    0.09087000611,  -0.1162228838,   0.5245254059,
       -0.02984068370,   0.2622627030,    0.08716716288,  0.4997850336,
       -0.1904774247,   -0.3424226677,    0.1961743797,  -0.006603146547,
       -0.009999592615,  0.003740870431,  0.02228152988, -0.1464888644,
        0.1282747834,    0.4253232566,   -0.6393164269,   0.03874458847,
       -0.3611982142,    0.4202921001,    0.2217097319,   0.2101460500,
        0.2708986607,    0.2130952366,   -0.1864672912,  -0.4589014084,
        0.3410943104,   -0.01528121247,  -0.03494751858,  0.006246941013
    ]

    d = dr.normalize(m.Array3f(1, 2, 3))

    for i in range(10):
        out = dr.sh_eval(d, i)
        assert dr.allclose(out, ref[:(i+1)**2], atol=5e-6)


@pytest.mark.parametrize("pkg", ['drjit.cuda', 'drjit.llvm'])
def test23_shape(pkg):
    m = get_class(pkg)

    assert dr.shape(m.Float()) == [0]
    assert dr.shape(m.Float(4.0)) == [1]
    assert dr.shape(m.Float([0, 1, 2])) == [3]

    assert dr.shape(m.Array3f(0, 1, 2)) == [3, 1]
    assert dr.shape(m.Array3f([0, 1], [2, 3], [4, 5])) == [3, 2]
    assert dr.shape(m.Array3f([0, 1], [2, 3], 4)) == [3, 2]

    assert dr.shape(m.Matrix3f([[0, 1, 2], [3, 4, 5], [6, 7, 8]])) == [3, 3, 1]
    assert dr.shape(m.Matrix3f([[0, 1, 2], [3, 4, 5], [6, 7, [8, 9]]])) == [3, 3, 2]

    # Ragged
    assert dr.shape(m.Array3f([0, 1, 2], [0, 1], [0, 1])) is None
    assert dr.shape(m.Array3f([0, 1], [0, 1, 2], [0, 1, 2, 3])) is None

@pytest.mark.parametrize("pkg", ['drjit.cuda.ad', 'drjit.llvm.ad'])
def test24_set_label(pkg):
    m = get_class(pkg)

    def check_label(v, label):
        assert dr.label(v) == label
        assert dr.label(dr.detach(v)) == dr.label(v)

    a = m.Float()
    check_label(a, None)
    dr.set_label(a, 'a')
    check_label(a, None)

    a = m.Float(4.0)
    b = m.Float(4.0)
    dr.set_label(b, 'b')
    check_label(a, None)
    check_label(b, 'b')
    c = m.Float(4.0)
    check_label(c, None)

    a = m.Array3f(4.0)
    check_label(a, [None, None, None])
    dr.set_label(a, 'a')
    check_label(a, ['a_0', 'a_1', 'a_2'])
    assert dr.label(dr.detach(a)) == ['a_0', 'a_1', 'a_2']


import math
import cmath
import drjit as dr
import pytest

classes = []

if dr.has_backend(dr.JitBackend.CUDA):
    classes.append(dr.cuda.Complex2f)
if dr.has_backend(dr.JitBackend.LLVM):
    classes.append(dr.llvm.Complex2f)

classes_dynamic = list(classes)

if hasattr(dr, 'packet'):
    classes.append(dr.packet.Complex2f)

@pytest.mark.parametrize('C', classes)
def test01_bcast(C):
    a = C()
    a.real = 5
    a.imag = 0
    b = C(C.Value(5))
    assert a == b


@pytest.mark.parametrize('C', classes)
def test02_sub(C):
    a = C(1, 3)
    b = C(5, 7)
    c = C(1-5, 3-7)
    assert a-b == c


@pytest.mark.parametrize('C', classes)
def test03_mul(C):
    a = C(1, 3)
    b = C(5, 7)
    c = a * b
    d = C(-16, 22)
    assert c == d
    assert a * 2 == C(2, 6)
    assert 2 * a == C(2, 6)


@pytest.mark.parametrize('C', classes)
def test04_div(C):
    a = C(1, 3)
    b = C(5, 7)
    c = a / b
    d = C(13 / 37, 4 / 37)
    assert dr.allclose(c, d)
    assert a / 2 == C(1/2, 3/2)


@pytest.mark.parametrize('C', classes)
def test05_rcp(C):
    assert dr.allclose(dr.rcp(C(1, 3)), C(1/10, -3/10))


@pytest.mark.parametrize('C', classes)
def test06_from_builtin(C):
    assert C(complex(3, 4)) == C(3, 4)


@pytest.mark.parametrize('C', classes)
def test07_from_builtin(C):
    # Fmadd should fallback to regular multiply (complex)-add
    assert dr.fma(C(2, 2), C(5, 5), C(5, 6)) == C(5, 26)


@pytest.mark.parametrize('C', classes)
def test08_misc(C):
    for i in range(-5, 5):
        for j in range(-5, 5):
            a = dr.sqrt(C(i, j))
            b = cmath.sqrt(complex(i, j))
            assert dr.allclose(a, C(b))

            assert dr.allclose(dr.conj(a), C(b.conjugate()))
            assert dr.allclose(dr.abs(a), C(abs(b)))

            if i != 0 and j != 0:
                a = dr.rsqrt(C(i, j))
                b = C(1 / cmath.sqrt(complex(i, j)))
                assert dr.allclose(a, b)


@pytest.mark.parametrize('C', classes)
def test09_trig(C):
    for i in range(-5, 5):
        for j in range(-5, 5):
            a = dr.sin(C(i, j))
            b = C(cmath.sin(complex(i, j)))
            assert dr.allclose(a, b)

            a = dr.cos(C(i, j))
            b = C(cmath.cos(complex(i, j)))
            assert dr.allclose(a, b)

            sa, ca = dr.sincos(C(i, j))
            sb = C(cmath.sin(complex(i, j)))
            cb = C(cmath.cos(complex(i, j)))
            assert dr.allclose(sa, sb)
            assert dr.allclose(ca, cb)

            # Python appears to handle the branch cuts
            # differently from DrJit, C, and Mathematica..
            a = dr.asin(C(i, j+0.1))
            b = C(cmath.asin(complex(i, j+0.1)))
            assert dr.allclose(a, b)

            a = dr.acos(C(i, j+0.1))
            b = C(cmath.acos(complex(i, j+0.1)))
            assert dr.allclose(a, b)

            if abs(j) != 1 or i != 0:
                a = dr.atan(C(i, j))
                b = C(cmath.atan(complex(i, j)))
                assert dr.allclose(a, b, atol=1e-7)


@pytest.mark.parametrize('C', classes)
def test10_math_explog(C):
    for i in range(-5, 5):
        for j in range(-5, 5):
            if i != 0 or j != 0:
                a = dr.log(C(i, j))
                b = C(cmath.log(complex(i, j)))
                assert dr.allclose(a, b)

                a = dr.log2(C(i, j))
                b = C(cmath.log(complex(i, j)) / cmath.log(2))
                assert dr.allclose(a, b)

            a = dr.exp(C(i, j))
            b = C(cmath.exp(complex(i, j)))
            assert dr.allclose(a, b)

            a = dr.exp2(C(i, j))
            b = C(cmath.exp(complex(i, j) * cmath.log(2)))
            assert dr.allclose(a, b)

            a = dr.power(C(2, 3), C(i, j))
            b = C(complex(2, 3) ** complex(i, j))
            assert dr.allclose(a, b)


@pytest.mark.parametrize('C', classes)
def test11_hyp(C):
    for i in range(-5, 5):
        for j in range(-5, 5):
            a = dr.sinh(C(i, j))
            b = C(cmath.sinh(complex(i, j)))
            assert dr.allclose(a, b)

            a = dr.cosh(C(i, j))
            b = C(cmath.cosh(complex(i, j)))
            assert dr.allclose(a, b)

            sa, ca = dr.sincosh(C(i, j))
            sb = C(cmath.sinh(complex(i, j)))
            cb = C(cmath.cosh(complex(i, j)))
            assert dr.allclose(sa, sb)
            assert dr.allclose(ca, cb)

            # Python appears to handle the branch cuts
            # differently from DrJit, C, and Mathematica..
            a = dr.asinh(C(i + 0.1, j))
            b = C(cmath.asinh(complex(i + 0.1, j)))
            assert dr.allclose(a, b)

            a = dr.acosh(C(i, j))
            b = C(cmath.acosh(complex(i, j)))
            assert dr.allclose(a, b, atol=1e-7)

            if abs(i) != 1 or j != 0:
                a = dr.atanh(C(i, j))
                b = C(cmath.atanh(complex(i, j)))
                assert dr.allclose(a, b, atol=1e-7)


@pytest.mark.parametrize('C', classes_dynamic)
def test12_numpy(C):
    pytest.importorskip("numpy")
    arr1 = C((1.0, 2.0, 3.0), (4.0, 5.0, 6.0))
    arr2 = arr1.numpy()
    assert 'j' in repr(arr2)
    arr3 = C(arr2)
    assert arr1 == arr3


@pytest.mark.parametrize('C', classes)
def test13_abs(C):
    assert dr.allclose(abs(C(1, 2)), math.sqrt(5))



import drjit as dr
import pytest
import importlib


@pytest.fixture(scope="module", params=['drjit.cuda.ad', 'drjit.llvm.ad'])
def m(request):
    if 'cuda' in request.param:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    else:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    yield importlib.import_module(request.param)


def test01_kernel_history(m):
    for i in range(4):
        dr.eval(dr.arange(m.Float, i + 4))

    # Kernel history should be disabled by default
    assert len(dr.kernel_history()) == 0

    assert not dr.flag(dr.JitFlag.KernelHistory)

    with dr.scoped_set_flag(dr.JitFlag.KernelHistory, True):
        assert dr.flag(dr.JitFlag.KernelHistory)
        for i in range(4):
            dr.eval(dr.arange(m.Float, i + 4))

        history = dr.kernel_history()
        assert len(history) == 4
        for i in range(4):
            assert history[i]['size'] == i + 4

    assert not dr.flag(dr.JitFlag.KernelHistory)

    # Kernel history should be erased after queried
    assert len(dr.kernel_history()) == 0



# TODO:
# - Check number of kernel launched when scheduling variables to make sure it create a single kernel
# - Check that number of output only contains the ones required (optimization)
# - ...



#!/usr/bin/env python
"""
Usage: generate_stub_files.py {dest_dir}

This script generates stub files for Python type information for the `drjit`
package.  It recursively traverses the `drjit` package and writes all the
objects (clases, methods, functions, enums, etc.) it finds to the `dest_dir`
folder. The stub files contain both the signatures and the docstrings of the
objects.
"""

import os
import inspect
import re
import sys
import logging

# ------------------------------------------------------------------------------

buffer = ''

def w(s):
    global buffer
    buffer += f'{s}\n'

# ------------------------------------------------------------------------------

def process_type_hint(s):
    sub = s
    type_hints = []
    offset = 0

    while True:
        match = re.search(r'[a-zA-Z]+: ', sub)
        if match is None:
            return s
        i = match.start() + len(match.group())
        match_next = re.search(r'[a-zA-Z]+: ', sub[i:])
        if match_next is None:
            j = sub.index(')')
        else:
            j = i + match_next.start() - 2

        type_hints.append((offset + i, offset + j, sub[i:j]))

        if match_next is None:
            break

        offset = offset + j
        sub = s[offset:]

    offset = 0
    result = ''
    for t in type_hints:
        result += s[offset:t[0]-2]
        offset = t[1]

        # if the type hint is valid, then add it as well
        if not ('::' in t[2]):
            result += f': {t[2]}'
    result += s[offset:]

    # Check is return type hint is not valid
    if '::' in result[result.index(' -> '):]:
        result = result[:result.index(' -> ')]

    return result

# ------------------------------------------------------------------------------

def process_properties(name, p, indent=0):
    indent = ' ' * indent

    if not p is None:
        w(f'{indent}{name} = ...')
        if not p.__doc__ is None:
            doc = p.__doc__.splitlines()
            if len(doc) == 1:
                w(f'{indent}\"{doc[0]}\"')
            elif len(doc) > 1:
                w(f'{indent}\"\"\"')
                for l in doc:
                    w(f'{indent}{l}')
                w(f'{indent}\"\"\"')

# ------------------------------------------------------------------------------

def process_enums(name, e, indent=0):
    indent = ' ' * indent

    if not e is None:
        w(f'{indent}{name} = {int(e)}')

        if not e.__doc__ is None:
            doc = e.__doc__.splitlines()
            w(f'{indent}\"\"\"')
            for l in doc:
                if l.startswith(f'  {name}'):
                    w(f'{indent}{l}')
            w(f'{indent}\"\"\"')

# ------------------------------------------------------------------------------

def process_class(obj):
    methods = []
    py_methods = []
    properties = []
    enums = []

    for k in dir(obj):
        # Skip private attributes
        if k.startswith('_'):
            continue

        if k.endswith('_'):
            continue

        v = getattr(obj, k)
        if type(v).__name__ == 'instancemethod':
            methods.append((k, v))
        elif type(v).__name__ == 'function' and v.__code__.co_varnames[0] == 'self':
            py_methods.append((k, v))
        elif type(v).__name__ == 'property':
            properties.append((k, v))
        elif str(v).endswith(k):
            enums.append((k, v))

    base = obj.__bases__[0]
    base_module = base.__module__
    base_name = base.__qualname__
    has_base = not (base_module == 'builtins' or base_name == 'object' or base_name == 'pybind11_object')
    base_name = base_module + '.' + base_name

    if has_base and not base.__module__.startswith('drjit'):
        w(f'from {base.__module__} import {base_name}')

    w(f'class {obj.__name__}{"(" + base_name + ")" if has_base else ""}:')
    if obj.__doc__ is not None:
        doc = obj.__doc__.splitlines()
        if len(doc) > 0:
            if doc[0].strip() == '':
                doc = doc[1:]
            if obj.__doc__:
                w(f'    \"\"\"')
                for l in doc:
                    w(f'    {l}')
                w(f'    \"\"\"')
                w(f'')

    process_function('__init__', obj.__init__, indent=4)
    process_function('__call__', obj.__call__, indent=4)

    if len(properties) > 0:
        for k, v in properties:
            process_properties(k, v, indent=4)
        w(f'')

    if len(enums) > 0:
        for k, v in enums:
            process_enums(k, v, indent=4)
        w(f'')

    for k, v in methods:
        process_function(k, v, indent=4)

    for k, v in py_methods:
        process_py_function(k, v, indent=4)

    w(f'    ...')
    w('')


# ------------------------------------------------------------------------------


def process_function(name, obj, indent=0):
    indent = ' ' * indent
    if obj is None or obj.__doc__ is None:
        return

    overloads = []
    for l in obj.__doc__.splitlines():
        if ') -> ' in l:
            l = process_type_hint(l)
            overloads.append((l, []))
        else:
            if len(overloads) > 0:
                overloads[-1][1].append(l)

    for l, doc in overloads:
        has_doc = len(doc) > 1

        # Overload?
        if l[1] == '.':
            w(f"{indent}@overload")
            w(f"{indent}def {l[3:]}:{'' if has_doc else ' ...'}")
        else:
            w(f"{indent}def {l}:{'' if has_doc else ' ...'}")

        if len(doc) > 1: # first line is always empty
            w(f'{indent}    \"\"\"')
            for l in doc[1:]:
                w(f'{indent}    {l}')
            w(f'{indent}    \"\"\"')
            w(f'{indent}    ...')
            w(f'')

# ------------------------------------------------------------------------------

def process_py_function(name, obj, indent=0):
    indent = ' ' * indent
    if obj is None:
        return

    has_doc = obj.__doc__ is not None

    signature = str(inspect.signature(obj))
    signature = signature.replace('\'', '')
    signature = signature.replace('dr.', 'drjit.')

    # Fix parameters that have enums as default values
    enum_match = re.search(r'\=<', signature)
    while enum_match is not None:
        begin = enum_match.start()
        end = begin + signature[begin:].index('>')

        new_default_value = signature[begin + 2:end]
        new_default_value = new_default_value[:new_default_value.index(':')]

        signature = signature[:begin + 1] + new_default_value + signature[end + 1:]
        enum_match = re.search(r'\=<', signature[begin])

    w(f"{indent}def {name}{signature}:{'' if has_doc else ' ...'}")

    if has_doc:
        doc = obj.__doc__.splitlines()
        if len(doc) > 0: # first line is always empty
            w(f'{indent}    \"\"\"')
            for l in doc:
                w(f'{indent}    {l.strip()}')
            w(f'{indent}    \"\"\"')
            w(f'{indent}    ...')
            w(f'')

# ------------------------------------------------------------------------------

def process_module(m, top_module=False):
    global buffer

    submodules = []
    buffer = ''

    w('from typing import Any, Callable, Iterable, Iterator, Tuple, List, TypeVar, Union, overload, ModuleType'
      )
    w('import drjit')
    w('import drjit as dr')
    w('')

    # Ignore initialization errors of invalid variants on this system
    try:
        dir(m)
    except Exception:
        pass

    for k in dir(m):
        v = getattr(m, k)

        if inspect.isclass(v):
            if (hasattr(v, '__module__')
                    and not v.__module__.startswith('drjit')):
                continue
            process_class(v)
        elif type(v).__name__ in ['method', 'function']:
            process_py_function(k, v)
        elif type(v).__name__ == 'builtin_function_or_method':
            process_function(k, v)
        elif type(v) in [str, bool, int, float]:
            if k.startswith('_'):
                continue
            process_properties(k, v)
        elif type(v).__bases__[0].__name__ == 'module' or type(v).__name__ == 'module':
            if k in ['dr']:
                continue
            if not v.__name__.startswith('drjit'):
                continue

            if k.startswith('_'):
                continue

            module_filename = v.__name__[len("drjit."):]
            module_filename = module_filename.replace('.', '_')

            w(f'')
            w(f'from {".stubs" if top_module else "."} import {module_filename} as {k}')
            w('')
            submodules.append((module_filename, v))

    return buffer, submodules

# ------------------------------------------------------------------------------

if __name__ == '__main__':
    logging.info('Generating stub files for the drjit package.')

    if len(sys.argv) < 2:
        raise RuntimeError("Exactly one argument is expected: the output "
                           "directory of the generated stubs")
    stub_folder = sys.argv[1]

    import drjit as dr

    os.makedirs(os.path.join(stub_folder, 'stubs'), exist_ok=True)

    logging.debug(f'Processing drjit root module')
    buffer, submodules = process_module(dr, top_module=True)
    with open(os.path.join(stub_folder, '__init__.pyi'), 'w') as f:
        f.write(buffer)

    processed_submodules = set()
    while len(submodules) != 0:
        k, v = submodules[0]
        if v in processed_submodules:
            submodules = submodules[1:]
            continue

        logging.debug(f'Processing submodule: {v.__name__}')
        buffer, new_submodules = process_module(v)

        with open(f'{os.path.join(stub_folder, "stubs", k + ".pyi")}', 'w') as f:
            f.write(buffer)

        submodules = submodules[1:] + new_submodules
        processed_submodules.add(v)

    logging.info(f'Done -> stub files written to {os.path.abspath(stub_folder)}')



# -*- coding: utf-8 -*-

from __future__ import print_function

import sys, re, os

try:
    from skbuild import setup
    import pybind11
except ImportError:
    print("The preferred way to invoke 'setup.py' is via pip, as in 'pip "
          "install .'. If you wish to run the setup script directly, you must "
          "first install the build dependencies listed in pyproject.toml!",
          file=sys.stderr)
    raise

VERSION_REGEX = re.compile(
    r"^\s*#\s*define\s+DRJIT_VERSION_([A-Z]+)\s+(.*)$", re.MULTILINE)

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join("include/drjit/fwd.h")) as f:
    matches = dict(VERSION_REGEX.findall(f.read()))
    drjit_version = "{MAJOR}.{MINOR}.{PATCH}".format(**matches)

with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

long_description = long_description[long_description.find('About this project'):]

drjit_cmake_toolchain_file = os.environ.get("DRJIT_CMAKE_TOOLCHAIN_FILE", "")

setup(
    name="drjit",
    version=drjit_version,
    author="Wenzel Jakob",
    author_email="wenzel.jakob@epfl.ch",
    description="A Just-In-Time-Compiler for Differentiable Rendering",
    url="https://github.com/mitsuba-renderer/drjit",
    license="BSD",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    cmake_args=[
        '-DDRJIT_ENABLE_JIT:BOOL=ON',
        '-DDRJIT_ENABLE_AUTODIFF:BOOL=ON',
        '-DDRJIT_ENABLE_PYTHON:BOOL=ON',
        '-DCMAKE_INSTALL_LIBDIR=drjit',
        '-DCMAKE_INSTALL_BINDIR=drjit',
        '-DCMAKE_INSTALL_INCLUDEDIR=drjit/include',
        '-DCMAKE_INSTALL_DATAROOTDIR=drjit/share',
        f'-DCMAKE_TOOLCHAIN_FILE={drjit_cmake_toolchain_file}'
    ],
    packages=['drjit'],
    python_requires=">=3.8"
)



import drjit as dr

import sys
import importlib
import pytest

try:
    import torch
except:
    pass

@pytest.fixture(scope="module", params=['drjit.cuda.ad', 'drjit.llvm.ad'])
def m(request):
    if not 'torch' in sys.modules:
        pytest.skip('PyTorch is not installed on this system')
    if 'cuda' in request.param:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    else:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    yield importlib.import_module(request.param)


def test01_to_torch_single_arg(m):
    a = m.TensorXf(m.Float([1.0, 2.0, 3.0]), shape=[3])
    dr.enable_grad(a)

    @dr.wrap_ad(source='drjit', target='torch')
    def func(a):
        return a * 4

    b = func(a)
    dr.backward(dr.sum(b))

    assert dr.allclose(b, [4, 8, 12])
    assert dr.allclose(dr.grad(a), [4, 4, 4])

    a = m.Float([1.0, 2.0, 3.0])
    dr.enable_grad(a)
    with pytest.raises(TypeError, match='should be Dr.Jit tensor'):
        b = func(a)


def test02_to_torch_two_args(m):
    a = m.TensorXf(m.Float([1.0, 2.0, 3.0]), shape=[3])
    b = m.TensorXf(m.Float([4.0, 5.0, 6.0]), shape=[3])
    dr.enable_grad(a, b)

    @dr.wrap_ad(source='drjit', target='torch')
    def func(a, b):
        return a * 4 + b * 3

    c = func(a, b)
    dr.backward(dr.sum(c))

    assert dr.allclose(c, [16, 23, 30])
    assert dr.allclose(dr.grad(a), [4, 4, 4])
    assert dr.allclose(dr.grad(b), [3, 3, 3])


def test03_to_torch_non_diff_arg_and_kwargs(m):
    a = m.TensorXf(m.Float([1.0, 2.0, 3.0]), shape=[3])
    dr.enable_grad(a)

    @dr.wrap_ad(source='drjit', target='torch')
    def func2(a, c: int = 4, s: str = '', d: int = 1):
        print(s)
        return a * c * d

    b = func2(a, s='test', c=4)
    dr.backward(dr.sum(b))

    assert dr.allclose(b, [4, 8, 12])
    assert dr.allclose(dr.grad(a), [4, 4, 4])


def test04_to_torch_two_args_two_outputs(m):
    a = m.TensorXf(m.Float([1.0, 2.0, 3.0]), shape=[3])
    b = m.TensorXf(m.Float([4.0, 5.0, 6.0]), shape=[3])
    dr.enable_grad(a, b)

    @dr.wrap_ad(source='drjit', target='torch')
    def func(a, b):
        return a * 4, b * 3

    c, d = func(a, b)
    dr.backward(dr.sum(c + d))

    assert dr.allclose(c, [4, 8, 12])
    assert dr.allclose(d, [12, 15, 18])
    assert dr.allclose(dr.grad(a), [4, 4, 4])
    assert dr.allclose(dr.grad(b), [3, 3, 3])


def test05_from_torch_single_arg(m):
    a = torch.tensor([1.0, 2.0, 3.0])
    if dr.is_cuda_v(m.Float):
        a = a.cuda()
    a.requires_grad = True

    @dr.wrap_ad(source='torch', target='drjit')
    def func(a):
        return a * 4

    b = func(a)

    b.sum().backward()

    assert dr.allclose(m.Float(b), [4, 8, 12])
    assert dr.allclose(m.Float(a.grad), [4, 4, 4])


def test06_from_torch_two_args(m):
    a = torch.tensor([1.0, 2.0, 3.0])
    b = torch.tensor([4.0, 5.0, 6.0])
    if dr.is_cuda_v(m.Float):
        a = a.cuda()
        b = b.cuda()
    a.requires_grad = True
    b.requires_grad = True

    @dr.wrap_ad(source='torch', target='drjit')
    def func(a, b):
        return a * 4 + b * 3

    c = func(a, b)

    c.sum().backward()

    assert dr.allclose(m.Float(c), [16, 23, 30])
    assert dr.allclose(m.Float(a.grad), [4, 4, 4])
    assert dr.allclose(m.Float(b.grad), [3, 3, 3])


def test07_from_torch_non_diff_args_and_kwargs(m):
    a = torch.tensor([1.0, 2.0, 3.0])
    if dr.is_cuda_v(m.Float):
        a = a.cuda()
    a.requires_grad = True

    @dr.wrap_ad(source='torch', target='drjit')
    def func2(a, c: int = 4, s: str = '', d: int = 1):
        print(s)
        return a * c * d

    b = func2(a, s='test', c=4)

    b.sum().backward()

    assert dr.allclose(m.Float(b), [4, 8, 12])
    assert dr.allclose(m.Float(a.grad), [4, 4, 4])


def test08_from_torch_two_args_two_outputs(m):
    a = torch.tensor([1.0, 2.0, 3.0])
    b = torch.tensor([4.0, 5.0, 6.0])
    if dr.is_cuda_v(m.Float):
        a = a.cuda()
        b = b.cuda()
    a.requires_grad = True
    b.requires_grad = True

    @dr.wrap_ad(source='torch', target='drjit')
    def func(a, b):
        return a * 4, b * 3

    c, d = func(a, b)

    e = (c + d).sum()
    e.backward()

    assert dr.allclose(m.Float(c), [4, 8, 12])
    assert dr.allclose(m.Float(d), [12, 15, 18])
    assert dr.allclose(m.Float(a.grad), [4, 4, 4])
    assert dr.allclose(m.Float(b.grad), [3, 3, 3])



from drjit import VarType, Exception, Dynamic
import sys as _sys
from collections.abc import Mapping as _Mapping, \
                            Sequence as _Sequence


def is_array_v(arg, /):
    '''
    is_array_v(arg, /)
    Check if the input is a Dr.Jit array instance or type

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` or type(``arg``) is a Dr.Jit array type, and ``False`` otherwise
    '''
    return getattr(arg, 'IsDrJit', False)


def size_v(arg, /):
    '''
    size_v(arg, /)
    Return the (static) size of the outermost dimension of the provided Dr.Jit
    array instance or type

    Note that this function mainly exists to query type-level information. Use the
    Python ``len()`` function to query the size in a way that does not distinguish
    between static and dynamic arrays.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        int: Returns either the static size or :py:data:`drjit.Dynamic` when
        ``arg`` is a dynamic Dr.Jit array. Returns ``1`` for all other types.
    '''
    return getattr(arg, 'Size', 1)


def depth_v(arg, /):
    '''
    depth_v(arg, /)
    Return the depth of the provided Dr.Jit array instance or type

    For example, an array consisting of floating point values (for example,
    :py:class:`drjit.scalar.Array3f`) has depth ``1``, while an array consisting of
    sub-arrays (e.g., :py:class:`drjit.cuda.Array3f`) has depth ``2``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        int: Returns the depth of the input, if it is a Dr.Jit array instance or
        type. Returns ``0`` for all other inputs.
    '''
    return getattr(arg, 'Depth', 0)


def scalar_t(arg, /):
    '''
    scalar_t(arg, /)
    Return the *scalar type* associated with the provided Dr.Jit array or type (i.e., the
    representation of elements at the lowest level)

    When the input is not a Dr.Jit array or type, the function returns the input
    unchanged. The following assertions illustrate the behavior of
    :py:func:`scalar_t`.


    .. code-block::

        assert dr.scalar_t(dr.scalar.Array3f) is bool
        assert dr.scalar_t(dr.cuda.Array3f) is float
        assert dr.scalar_t(dr.cuda.Matrix4f) is float
        assert dr.scalar_t("test") is str

    Args:
        arg (object): An arbitrary Python object

    Returns:
        int: Returns the scalar type of the provided Dr.Jit array, or the type of
        the input.
    '''
    if not isinstance(arg, type):
        arg = type(arg)
    return getattr(arg, 'Scalar', arg)


def value_t(arg, /):
    '''
    value_t(arg, /)
    Return the *value type* underlying the provided Dr.Jit array or type (i.e., the
    type of values obtained by accessing the contents using a 1D index).

    When the input is not a Dr.Jit array or type, the function returns the input
    unchanged. The following code fragment shows several example uses of
    :py:func:`value_t`.

    .. code-block::

        assert dr.value_t(dr.scalar.Array3f) is float
        assert dr.value_t(dr.cuda.Array3f) is dr.cuda.Float
        assert dr.value_t(dr.cuda.Matrix4f) is dr.cuda.Array4f
        assert dr.value_t(dr.cuda.TensorXf) is float
        assert dr.value_t("test") is str

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Returns the value type of the provided Dr.Jit array, or the type of
        the input.
    '''
    if not isinstance(arg, type):
        arg = type(arg)
    return getattr(arg, 'Value', arg)


def mask_t(arg, /):
    '''
    mask_t(arg, /)
    Return the *mask type* associated with the provided Dr.Jit array or type (i.e., the
    type produced by comparisons involving the argument).

    When the input is not a Dr.Jit array or type, the function returns the scalar
    Python ``bool`` type. The following assertions illustrate the behavior of
    :py:func:`mask_t`.


    .. code-block::

        assert dr.mask_t(dr.scalar.Array3f) is dr.scalar.Array3b
        assert dr.mask_t(dr.cuda.Array3f) is dr.cuda.Array3b
        assert dr.mask_t(dr.cuda.Matrix4f) is dr.cuda.Array44b
        assert dr.mask_t("test") is bool

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Returns the mask type associated with the input or ``bool`` when the
        input is not a Dr.Jit array.
    '''
    if not isinstance(arg, type):
        arg = type(arg)
    return getattr(arg, 'MaskType', bool)


def is_mask_v(arg, /):
    '''
    is_mask_v(arg, /)
    Check whether the input array instance or type is a Dr.Jit mask array or a
    Python ``bool`` value/type.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents a Dr.Jit mask array or Python ``bool``
        instance or type.
    '''
    return scalar_t(arg) is bool


def is_float_v(arg, /):
    '''
    is_float_v(arg, /)
    Check whether the input array instance or type is a Dr.Jit floating point array
    or a Python ``float`` value/type.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents a Dr.Jit floating point array or
        Python ``float`` instance or type.
    '''
    return scalar_t(arg) is float


def is_integral_v(arg, /):
    '''
    is_integral_v(arg, /)
    Check whether the input array instance or type is an integral Dr.Jit array
    or a Python ``int`` value/type.

    Note that a mask array is not considered to be integral.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an integral Dr.Jit array or
        Python ``int`` instance or type.
    '''
    return scalar_t(arg) is int


def is_arithmetic_v(arg, /):
    '''
    is_arithmetic_v(arg, /)
    Check whether the input array instance or type is an arithmetic Dr.Jit array
    or a Python ``int`` or ``float`` value/type.

    Note that a mask type (e.g. ``bool``, :py:class:`drjit.scalar.Array2b`, etc.)
    is *not* considered to be arithmetic.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an arithmetic Dr.Jit array or
        Python ``int`` or ``float`` instance or type.
    '''
    return scalar_t(arg) in [int, float]


def is_cuda_v(arg, /):
    '''
    is_cuda_v(arg, /)
    Check whether the input is a Dr.Jit CUDA array instance or type.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an array type from the
        ``drjit.cuda.*`` namespace, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsCUDA', False)


def is_llvm_v(arg, /):
    '''
    is_llvm_v(arg, /)
    Check whether the input is a Dr.Jit LLVM array instance or type.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an array type from the
        ``drjit.llvm.*`` namespace, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsLLVM', False)


def is_jit_v(arg, /):
    '''
    is_jit_v(arg, /)
    Check whether the input array instance or type represents a type that
    undergoes just-in-time compilation.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an array type from the
        ``drjit.cuda.*`` or ``drjit.llvm.*`` namespaces, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsJIT', False)


def is_diff_v(arg, /):
    '''
    is_diff_v(arg, /)
    Check whether the input is a differentiable Dr.Jit array instance or type.

    Note that this is a type-based statement that is unrelated to mathematical
    differentiability. For example, the integral type :py:class:`drjit.cuda.ad.Int`
    from the CUDA AD namespace satisfies ``is_diff_v(..) = 1``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an array type from the
        ``drjit.[cuda/llvm].ad.*`` namespace, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsDiff', False)


def is_complex_v(arg, /):
    '''
    is_complex_v(arg, /)
    Check whether the input is a Dr.Jit array instance or type representing a complex number.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if the test was successful, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsComplex', False) or isinstance(arg, complex)


def is_matrix_v(arg, /):
    '''
    is_matrix_v(arg, /)
    Check whether the input is a Dr.Jit array instance or type representing a matrix.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if the test was successful, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsMatrix', False)


def is_quaternion_v(arg, /):
    '''
    is_quaternion_v(arg, /)
    Check whether the input is a Dr.Jit array instance or type representing a quaternion.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if the test was successful, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsQuaternion', False)


def is_tensor_v(arg, /):
    '''
    is_tensor_v(arg, /)
    Check whether the input is a Dr.Jit array instance or type representing a tensor.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if the test was successful, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsTensor', False)


def is_texture_v(a):
    return getattr(a, 'IsTexture', False)


def is_vector_v(a):
    return getattr(a, 'IsVector', False)


def is_special_v(arg, /):
    '''
    is_special_v(arg, /)
    Check whether the input is a *special* Dr.Jit array instance or type.

    A *special* array type requires precautions when performing arithmetic
    operations like multiplications (complex numbers, quaternions, matrices).

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if the test was successful, and ``False`` otherwise.
    '''
    return getattr(arg, 'IsSpecial', False)


def is_static_array_v(a):
    return getattr(a, 'Size', Dynamic) != Dynamic


def is_dynamic_array_v(a):
    return getattr(a, 'Size', Dynamic) == Dynamic


def is_dynamic_v(a):
    return getattr(a, 'IsDynamic', False)


def is_unsigned_v(arg, /):
    '''
    is_unsigned_v(arg, /)
    Check whether the input array instance or type is an unsigned integer Dr.Jit
    array or a Python ``bool`` value/type (masks and boolean values are also
    considered to be unsigned).

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an unsigned Dr.Jit array or
        Python ``bool`` instance or type.
    '''
    if not is_array_v(arg):
        return False

    vt = arg.Type

    return vt == VarType.UInt8 or \
        vt == VarType.UInt16 or \
        vt == VarType.UInt32 or \
        vt == VarType.UInt64


def is_signed_v(arg, /):
    '''
    is_signed_v(arg, /)
    Check whether the input array instance or type is an signed Dr.Jit array
    or a Python ``int`` or ``float`` value/type.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` represents an signed Dr.Jit array or
        Python ``int`` or ``float`` instance or type.
    '''
    return not is_unsigned_v(arg)


def is_iterable_v(a):
    if isinstance(a, str):
        return False
    try:
        iter(a)
        return True
    except TypeError:
        return False


def bool_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *boolean* version with
    the same element size.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3i64`), it
       returns an *boolean* version (e.g. :py:class:`drjit.cuda.Array3b64`).

    2. When the input isn't a type, it returns ``bool_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``bool``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return mask_t(arg)

def int_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *signed integer*
    version with the same element size.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f64`), it
       returns an *signed integer* version (e.g. :py:class:`drjit.cuda.Array3u64`).

    2. When the input isn't a type, it returns ``int_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    if not is_array_v(arg):
        return int

    size = arg.Type.Size
    if size == 1:
        vt = VarType.Int8
    elif size == 2:
        vt = VarType.Int16
    elif size == 4:
        vt = VarType.Int32
    elif size == 8:
        vt = VarType.Int64
    else:
        raise Exception("Unsupported variable size!")

    t = arg.ReplaceScalar(vt)
    return t


def uint_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *unsigned integer*
    version with the same element size.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f64`), it
       returns an *unsigned integer* version (e.g. :py:class:`drjit.cuda.Array3u64`).

    2. When the input isn't a type, it returns ``uint_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    if not is_array_v(arg):
        return int

    size = arg.Type.Size
    if size == 1:
        vt = VarType.UInt8
    elif size == 2:
        vt = VarType.UInt16
    elif size == 4:
        vt = VarType.UInt32
    elif size == 8:
        vt = VarType.UInt64
    else:
        raise Exception("Unsupported variable size!")

    t = arg.ReplaceScalar(vt)
    return t


def float_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *floating point*
    version with the same element size.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3u64`), it
       returns an *floating point* version (e.g. :py:class:`drjit.cuda.Array3f64`).

    2. When the input isn't a type, it returns ``float_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``float``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    if not is_array_v(arg):
        return int

    size = arg.Type.Size
    if size == 2:
        vt = VarType.Float16
    elif size == 4:
        vt = VarType.Float32
    elif size == 8:
        vt = VarType.Float64
    else:
        raise Exception("Unsupported variable size!")

    return arg.ReplaceScalar(vt)


def uint32_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *unsigned 32 bit*
    version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f`), it
       returns an *unsigned 32 bit* version (e.g. :py:class:`drjit.cuda.Array3u`).

    2. When the input isn't a type, it returns ``uint32_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.UInt32) if is_array_v(arg) else int


def int32_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into a *signed 32 bit*
    version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f`), it
       returns an *signed 32 bit* version (e.g. :py:class:`drjit.cuda.Array3i`).

    2. When the input isn't a type, it returns ``int32_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.Int32) if is_array_v(arg) else int


def uint64_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into an *unsigned 64 bit*
    version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f`), it
       returns an *unsigned 64 bit* version (e.g. :py:class:`drjit.cuda.Array3u64`).

    2. When the input isn't a type, it returns ``uint64_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.UInt64) if is_array_v(arg) else int


def int64_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into an *signed 64 bit* version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f`), it
       returns an *signed 64 bit* version (e.g. :py:class:`drjit.cuda.Array3i64`).

    2. When the input isn't a type, it returns ``int64_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``int``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.Int64) if is_array_v(arg) else int


def float32_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into an 32 bit floating point version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3u`), it
       returns a *32 bit floating point* version (e.g. :py:class:`drjit.cuda.Array3f`).

    2. When the input isn't a type, it returns ``float32_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``float``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.Float32) if is_array_v(arg) else float


def float64_array_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into an 64 bit floating point version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3u`), it
       returns a *64 bit floating point* version (e.g. :py:class:`drjit.cuda.Array3f64`).

    2. When the input isn't a type, it returns ``float64_array_t(type(arg))``.

    3. When the input is not a Dr.Jit array or type, the function returns ``float``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    return arg.ReplaceScalar(VarType.Float64) if is_array_v(arg) else float


def diff_array_t(a):
    '''
    Converts the provided Dr.Jit array/tensor type into a differentiable version.

    This function implements the following set of behaviors:

    1. When invoked with a Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.Array3f`), it
       returns a *differentiable* version (e.g. :py:class:`drjit.cuda.ad.Array3f`).

    2. When the input isn't a type, it returns ``diff_array_t(type(arg))``.

    3. When the input is is a list or a tuple, it recursively call ``diff_array_t`` over all elements.

    4. When the input is not a Dr.Jit array or type, the function throws an exception.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    if isinstance(a, tuple):
        return tuple(diff_array_t(v) for v in a)
    elif isinstance(a, list):
        return [diff_array_t(v) for v in a]
    elif not isinstance(a, type):
        return diff_array_t(type(a))
    elif not is_array_v(a):
        raise Exception("diff_array_t(): requires an Dr.Jit input array!")
    elif a.IsDiff:
        return a
    else:
        return a.ReplaceScalar(a.Type, diff=True)


def detached_t(arg):
    '''
    Converts the provided Dr.Jit array/tensor type into an non-differentiable version.

    This function implements the following set of behaviors:

    1. When invoked with a differentiable Dr.Jit array *type* (e.g. :py:class:`drjit.cuda.ad.Array3f`), it
       returns a non-differentiable version (e.g. :py:class:`drjit.cuda.Array3f`).

    2. When the input isn't a type, it returns ``detached_t(type(arg))``.

    3. When the input type is non-differentiable or not a Dr.Jit array type, the function returns it unchanged.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the conversion as described above.
    '''
    if not is_array_v(arg):
        raise Exception("detached_t(): requires an Dr.Jit input array!")
    elif not isinstance(arg, type):
        return detached_t(type(arg))
    elif not arg.IsDiff:
        return arg
    else:
        return arg.ReplaceScalar(arg.Type, diff=False)


def is_struct_v(arg, /):
    '''
    is_struct_v(arg, /)
    Check if the input is a Dr.Jit-compatible data structure

    Custom data structures can be made compatible with various Dr.Jit operations by
    specifying a ``DRJIT_STRUCT`` member. See the section on :ref:`custom data
    structure <custom-struct>` for details. This type trait can be used to check
    for the existence of such a field.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        bool: ``True`` if ``arg`` has a ``DRJIT_STRUCT`` member
    '''
    return hasattr(arg, 'DRJIT_STRUCT')


def leaf_array_t(arg):
    '''
    Extracts a leaf array type underlying a Python object tree, with a preference
    for differentiable arrays.

    This function implements the following set of behaviors:

    1. When the input isn't a type, it returns ``leaf_array_t(type(arg))``.

    2. When invoked with a Dr.Jit array type, returns the lowest-level array type
       underlying a potentially nested array.

    3. When invoked with a sequence, mapping or custom data structure made of Dr.Jit arrays,
       examines underlying Dr.Jit array types and returns the lowest-level array type with
       a preference for differentiable arrays and floating points arrays.
       E.g. when passing a list containing arrays of type :py:class:`drjit.cuda.ad.Float` and :py:class:`drjit.cuda.UInt`,
       the function will return :py:class:`drjit.cuda.ad.Float`.

    4. Otherwise returns ``None``.

    Args:
        arg (object): An arbitrary Python object

    Returns:
        type: Result of the extraction as described above.
    '''
    t = None

    if isinstance(arg, _Sequence):
        for e in arg:
            t = leaf_array_t(e)
            if is_diff_v(t) and is_float_v(t):
                break
    elif isinstance(arg, _Mapping):
        for k, v in arg:
            t = leaf_array_t(v)
            if is_diff_v(t) and is_float_v(t):
                break
    elif is_struct_v(arg):
        for k in type(arg).DRJIT_STRUCT.keys():
            t = leaf_array_t(getattr(arg, k))
            if is_diff_v(t) and is_float_v(t):
                break
    elif is_tensor_v(arg):
        t = leaf_array_t(arg.Array)
    elif is_array_v(arg):
        t = arg
        if not isinstance(t, type):
            t = type(t)
        while is_array_v(value_t(t)):
            t = t.Value

    return t



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + "/build")

import drjit

# -- Project information -----------------------------------------------------

project = 'drjit'
copyright = '2022, Realistic Graphics Lab'
author = 'Realistic Graphics Lab'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinxcontrib.katex',
    'sphinx_togglebutton'
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# -- Options for HTML output -------------------------------------------------

html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']
html_css_files = ['custom.css']



import torch

from pathlib import Path
import numpy as np
from PIL import Image as im
import imageio.v3 as iio
import torch

def write_exr(exr_path, image):
    image = to_numpy(image).astype(np.float32)
    if len(image.shape) == 3:
        image = np.expand_dims(image, axis=0)
    try:
        iio.imwrite(exr_path, image)
    except OSError:
        imageio.plugins.freeimage.download()
        iio.imwrite(exr_path, image, extension='.exr')


def to_numpy(data):
    if torch.is_tensor(data):
        return data.detach().cpu().numpy()
    else:
        return np.array(data)

def linear_to_srgb(l):
    s = np.zeros_like(l)
    m = l <= 0.00313066844250063
    s[m] = l[m] * 12.92
    s[~m] = 1.055*(l[~m]**(1.0/2.4))-0.055
    return s


def srgb_to_linear(s):
    l = np.zeros_like(s)
    m = s <= 0.0404482362771082
    l[m] = s[m] / 12.92
    l[~m] = ((s[~m]+0.055)/1.055) ** 2.4
    return l

def to_srgb(image):
    image = to_numpy(image)
    if image.shape[2] == 4:
        image_alpha = image[:, :, 3:4]
        image = linear_to_srgb(image[:, :, 0:3])
        image = np.concatenate([image, image_alpha], axis=2)
    else:
        image = linear_to_srgb(image)
    return np.clip(image, 0, 1)

def write_jpg(jpg_path, image):
    image = to_srgb(to_numpy(image))
    image = (image * 255).astype(np.uint8)
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)
    rgb_im = im.fromarray(image).convert('RGB')
    rgb_im.save(jpg_path, format='JPEG', quality=95)
    
def read_image(image_path, is_srgb=None, remove_alpha=True):
    image_path = Path(image_path)
    image = iio.imread(image_path)
    image = np.atleast_3d(image)
    if remove_alpha and image.shape[2] == 4:
        image = image[:, :, 0:3]

    if image.dtype == np.uint8 or image.dtype == np.int16:
        image = image.astype("float32") / 255.0
    elif image.dtype == np.uint16 or image.dtype == np.int32:
        image = image.astype("float32") / 65535.0

    if is_srgb is None:
        if image_path.suffix in ['.exr', '.hdr', '.rgbe']:
            is_srgb = False
        else:
            is_srgb = True

    if is_srgb:
        image = to_linear(image)

    return image

def write_image(image_path, image, is_srgb=None):
    image_path = Path(image_path)
    image = to_numpy(image)
    image = np.atleast_3d(image)
    if image.shape[2] == 1:
        image = np.repeat(image, 3, axis=2)

    if is_srgb is None:
        if image_path.suffix in ['.exr', '.hdr', '.rgbe']:
            is_srgb = False
        else:
            is_srgb = True

    if is_srgb:
        image = to_srgb(image)

    if image_path.suffix == '.exr':
        image = image.astype(np.float32)
    else:
        image = (image * 255).astype(np.uint8)

    iio.imwrite(image_path, image)



import drjit as dr
import pytest
import importlib

@pytest.fixture(scope="module", params=['drjit.cuda', 'drjit.llvm'])
def m(request):
    if 'cuda' in request.param:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    else:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    yield importlib.import_module(request.param)


def test01_sum(m):
    assert dr.allclose(dr.sum(6.0), 6)

    a = dr.sum(m.Float([1, 2, 3]))
    assert dr.allclose(a, 6)
    assert type(a) is m.Float

    a = dr.sum([1.0, 2.0, 3.0])
    assert dr.allclose(a, 6)
    assert type(a) is float

    a = dr.sum(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Float

    a = dr.sum(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Int

    a = dr.sum(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, [6, 9, 12])
    assert type(a) is m.Float

    a = dr.sum_nested(m.Array3f([1, 1], [2, 2], [3, 3]))
    assert dr.allclose(a, 12)
    assert type(a) is m.Float

    a = dr.sum_nested(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Int

    a = dr.sum_nested(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, 27)
    assert type(a) is m.Float


def test02_mean(m):
    assert dr.allclose(dr.sum(6.0), 6)

    a = dr.mean(m.Float([1, 2, 3]))
    assert dr.allclose(a, 6 / 3.0)
    assert type(a) is m.Float

    a = dr.mean([1.0, 2.0, 3.0])
    assert dr.allclose(a, 6 / 3.0)
    assert type(a) is float

    a = dr.mean(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 6 / 3.0)
    assert type(a) is m.Float

    a = dr.mean(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6 / 3.0)
    assert type(a) is m.Float

    a = dr.mean(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, [2, 3, 4])
    assert type(a) is m.Float

    a = dr.mean_nested(m.Array3f([1, 1], [2, 2], [3, 3]))
    assert dr.allclose(a, 12 / 6.0)
    assert type(a) is m.Float

    a = dr.mean_nested(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6 / 3.0)
    assert type(a) is m.Float

    a = dr.mean_nested(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, 27 / 9.0)
    assert type(a) is m.Float


def test03_prod(m):
    assert dr.allclose(dr.prod(6.0), 6.0)

    a = dr.prod(m.Float([1, 2, 3]))
    assert dr.allclose(a, 6)
    assert type(a) is m.Float

    a = dr.prod([1.0, 2.0, 3.0])
    assert dr.allclose(a, 6)
    assert type(a) is float

    a = dr.prod(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Float

    a = dr.prod(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Int

    a = dr.prod(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, [6, 24, 60])
    assert type(a) is m.Float


    a = dr.prod_nested(m.Array3f([1, 1], [2, 2], [3, 3]))
    assert dr.allclose(a, 6 * 6)
    assert type(a) is m.Float

    a = dr.prod_nested(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 6)
    assert type(a) is m.Int

    a = dr.prod_nested(m.ArrayXf([1, 2, 3], [2, 3, 4], [3, 4, 5]))
    assert dr.allclose(a, 6 * 24 * 60)
    assert type(a) is m.Float


def test04_max(m):
    assert dr.allclose(dr.max(6.0), 6.0)

    a = dr.max(m.Float([1, 2, 3]))
    assert dr.allclose(a, 3)
    assert type(a) is m.Float

    a = dr.max([1.0, 2.0, 3.0])
    assert dr.allclose(a, 3)
    assert type(a) is float

    a = dr.max(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 3)
    assert type(a) is m.Float

    a = dr.max(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 3)
    assert type(a) is m.Int

    a = dr.max(m.ArrayXf([1, 2, 5], [2, 3, 4], [3, 4, 3]))
    assert dr.allclose(a, [3, 4, 5])
    assert type(a) is m.Float

    a = dr.max_nested(m.Array3f([1, 1], [2, 4], [3, 3]))
    assert dr.allclose(a, 4)
    assert type(a) is m.Float

    a = dr.max_nested(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 3)
    assert type(a) is m.Int

    a = dr.max_nested(m.ArrayXf([1, 2, 5], [2, 3, 4], [3, 4, 3]))
    assert dr.allclose(a, 5)
    assert type(a) is m.Float


def test05_min(m):
    assert dr.allclose(dr.min(6.0), 6.0)

    a = dr.min(m.Float([1, 2, 3]))
    assert dr.allclose(a, 1)
    assert type(a) is m.Float

    a = dr.min([1.0, 2.0, 3.0])
    assert dr.allclose(a, 1)
    assert type(a) is float

    a = dr.min(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 1)
    assert type(a) is m.Float

    a = dr.min(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 1)
    assert type(a) is m.Int

    a = dr.min(m.ArrayXf([1, 2, 5], [2, 3, 4], [3, 4, 3]))
    assert dr.allclose(a, [1, 2, 3])
    assert type(a) is m.Float

    a = dr.min_nested(m.Array3f([1, 1], [2, 4], [3, 3]))
    assert dr.allclose(a, 1)
    assert type(a) is m.Float

    a = dr.min_nested(m.Array3i(1, 2, 3))
    assert dr.allclose(a, 1)
    assert type(a) is m.Int

    a = dr.min_nested(m.ArrayXf([1, 2, 5], [2, 3, 4], [3, 4, 3]))
    assert dr.allclose(a, 1)
    assert type(a) is m.Float


def test06_dot(m):
    a = dr.dot(m.Float(1, 2, 3), m.Float(2, 1, 1))
    assert dr.allclose(a, 7)
    assert type(a) is m.Float

    a = dr.dot(m.Float(1, 2, 3), m.Float(2))
    assert dr.allclose(a, 12)
    assert type(a) is m.Float

    a = dr.dot(m.Float(1, 2, 3), [2, 1, 1])
    assert dr.allclose(a, 7)
    assert type(a) is m.Float

    a = dr.dot(m.Array3f(1, 2, 3), m.Array3f([2, 1, 1]))
    assert dr.allclose(a, 7)
    assert type(a) is m.Float

    a = dr.dot(m.ArrayXf([1, 2, 5], [2, 1, 4], [3, 4, 3]), m.ArrayXf([[1, 2, 5], [2, 2, 4], [3, 4, 3]]))
    assert dr.allclose(a, [14, 22, 50])
    assert type(a) is m.Float


def test07_norm(m):
    a = dr.norm(m.Float(1, 2, 3))
    assert dr.allclose(a, 3.74166)
    assert type(a) is m.Float

    a = dr.norm(m.Array3f(1, 2, 3))
    assert dr.allclose(a, 3.74166)
    assert type(a) is m.Float



import drjit as dr
import pytest
import importlib


def prepare(pkg):
    if 'cuda' in pkg:
        if not dr.has_backend(dr.JitBackend.CUDA):
            pytest.skip('CUDA mode is unsupported')
    elif 'llvm' in pkg:
        if not dr.has_backend(dr.JitBackend.LLVM):
            pytest.skip('LLVM mode is unsupported')
    return importlib.import_module(pkg)


@pytest.mark.parametrize("package", ['drjit.cuda', 'drjit.cuda.ad',
                                     'drjit.llvm', 'drjit.llvm.ad'])
def test_zero_initialization(package):
    package = prepare(package)
    Float, Array3f = package.Float, package.Array3f

    class MyStruct:
        DRJIT_STRUCT = { 'a' : Array3f, 'b' : Float }

        def __init__(self):
            self.a = Array3f()
            self.b = Float()

        # Custom zero initialize callback
        def zero_(self, size):
            self.a += 1

    foo = dr.zeros(MyStruct, 4)
    assert dr.width(foo) == 4
    assert foo.a == 1
    assert foo.b == 0

    foo = dr.zeros(MyStruct, 1)
    dr.resize(foo, 8)
    assert dr.width(foo) == 8


@pytest.mark.parametrize("package", ['drjit.cuda.ad', 'drjit.llvm.ad'])
def test_ad_operations(package):
    package = prepare(package)
    Float, Array3f = package.Float, package.Array3f

    class MyStruct:
        DRJIT_STRUCT = { 'a' : Array3f, 'b' : Float }

        def __init__(self):
            self.a = Array3f()
            self.b = Float()

    foo = dr.zeros(MyStruct, 4)
    assert not dr.grad_enabled(foo.a)
    assert not dr.grad_enabled(foo.b)
    assert not dr.grad_enabled(foo)

    dr.enable_grad(foo)
    assert dr.grad_enabled(foo.a)
    assert dr.grad_enabled(foo.b)
    assert dr.grad_enabled(foo)

    foo_detached = dr.detach(foo)
    assert not dr.grad_enabled(foo_detached.a)
    assert not dr.grad_enabled(foo_detached.b)
    assert not dr.grad_enabled(foo_detached)

    x = Float(4.0)
    dr.enable_grad(x)
    foo.a += x
    foo.b += x*x
    dr.forward(x)
    foo_grad = dr.grad(foo)
    assert foo_grad.a == 1
    assert foo_grad.b == 8

    dr.set_grad(foo, 5.0)
    foo_grad = dr.grad(foo)
    assert foo_grad.a == 5.0
    assert foo_grad.b == 5.0

    dr.accum_grad(foo, 5.0)
    foo_grad = dr.grad(foo)
    assert foo_grad.a == 10.0
    assert foo_grad.b == 10.0



###############################################################################
# GDB Script to improve introspection of array types when debugging software
# using DrJit. Copy this file to "~/.gdb" (creating the directory, if not
# present) and then apppend the following line to the file "~/.gdbinit"
# (again, creating it if, not already present):
###############################################################################
# set print pretty
# source ~/.gdb/drjit_gdb.py
###############################################################################

import gdb

simple_types = {
    'bool',
    'char', 'unsigned char',
    'short', 'unsigned short',
    'int', 'unsigned int',
    'long', 'unsigned long',
    'long long', 'unsigned long long',
    'float', 'double'
}


class DrJitIterator:
    def __init__(self, instance, size):
        self.instance = instance
        self.size = size
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= self.size:
            raise StopIteration
        result = ('[%i]' % self.index, self.instance[self.index])
        self.index += 1
        return result

    def next(self):
        return self.__next__()


class DrJitStaticArrayPrinter:
    def __init__(self, instance):
        self.instance = instance
        itype = self.instance.type.strip_typedefs()

        # Extract derived type
        if 'StaticArrayImpl' in str(itype):
            itype = itype.template_argument(3)

        try:
            data = self.instance['m_data']['_M_elems']
            self.data_type = data.type.strip_typedefs().target()
        except Exception:
            self.data_type = itype.template_argument(0)

        # Determine the size and data type
        self.size = int(str(itype.template_argument(1)))
        self.is_simple = str(self.data_type) in simple_types
        self.type_size = self.data_type.sizeof
        self.is_mask = 'Mask' in str(itype)

        try:
            _ = instance['k']
            self.kmask = True
        except Exception:
            self.kmask = False

    def entry(self, i):
        if i < 0 or i >= self.size:
            return None
        addr = int(self.instance.address) + self.type_size * i
        cmd = '*((%s *) 0x%x)' % (str(self.data_type), addr)
        return str(gdb.parse_and_eval(cmd))

    def children(self):
        if self.is_simple:
            return []
        else:
            return DrJitIterator(self.instance['m_data']['_M_elems'], self.size)

    def to_string(self):
        if self.is_simple:
            if not self.is_mask:
                result = [self.entry(i) for i in range(self.size)]
            else:
                if self.kmask:
                    # AVX512 mask register
                    result = list(reversed(format(int(self.instance['k']), '0%ib' % self.size)))
                else:
                    result = [None] * self.size
                    for i in range(self.size):
                        value = self.entry(i)
                        result[i] = '0' if (value == '0' or value == 'false') else '1'
            return '[' + ', '.join(result) + ']'
        else:
            return ''


class DrJitDynamicArrayPrinter:
    def __init__(self, instance):
        self.instance = instance
        itype = self.instance.type.strip_typedefs()
        self.size = int(str(self.instance['m_size']))
        self.packet_count = int(str(self.instance['m_packets_allocated']))
        self.packet_type = itype.template_argument(0)
        self.packet_size = self.packet_type.sizeof
        self.data = int(str(instance['m_packets']['_M_t']['_M_t']['_M_head_impl']), 0)
        self.limit = 20

    def to_string(self):
        values = []
        for i in range(self.packet_count):
            addr = int(self.data) + self.packet_size * i
            cmd = '*((%s *) 0x%x)' % (str(self.packet_type), addr)
            value = str(gdb.parse_and_eval(cmd))
            assert value[-1] == ']'
            values += value[value.rfind('[')+1:-1].split(', ')
            if len(values) > self.size:
                values = values[0:self.size]
                break
            if len(values) > self.limit:
                break
        if len(values) > self.limit:
            values = values[0:self.limit]
            values.append(".. %i skipped .." % (self.size - self.limit))
        return '[' + ', '.join(values) + ']'

# Static DrJit arrays
regexp_1 = r'(drjit::)?(Array|Packet|Complex|Matrix|' \
    'Quaternion|StaticArrayImpl)(Mask)?<.+>'

# Mitsuba 2 is one of the main users of DrJit. For convenience, also
# declare its custom array types here
regexp_2 = r'(mitsuba::)?(Vector|Point|Normal|Spectrum|Color)<.+>'

regexp_combined = r'^(%s)|(%s)$' % (regexp_1, regexp_2)

p = gdb.printing.RegexpCollectionPrettyPrinter("drjit")
p.add_printer("static", regexp_combined, DrJitStaticArrayPrinter)
p.add_printer("dynamic", r'^(drjit::)?DynamicArray(Impl)?<.+>$', DrJitDynamicArrayPrinter)

o = gdb.current_objfile()
gdb.printing.register_pretty_printer(o, p)
